#  VGI - VHDL Graphical Interface
#
#  Copyright: Stefan Doll (GNU License applies)
#
#  Function: Filter which creates a VHDL netlist from
#            a Tgif file, can also extract FSMs
#
#  Revision History:
#  -----------------
#  V 1.0 :  Preliminary
#  V 1.1 :  1st Beta Test
#  V 1.2 :  2nd Beta Test
#  V 1.3 :  Added init symbol to assign values to generic pins
#  V 1.6 :  replaced \' with ' in regular expressions 
#           (caused a problem with gawk-3.0.0)
#  V 1.7 :  added support for FSMs
#
#
# Problems:
#  * is it necessary to look for matching endpoints at all?
#    (should be covered by point_on_segment)
# 


# find a defined object:
#   sym  : entity declaration
#   icon : component instantiation
#   poly : signal
#

BEGIN		{object="none"}
BEGIN		{attribute="none"}
BEGIN		{portnum=0}
BEGIN		{gportnum=0}
BEGIN		{comnum=0}
BEGIN		{labelnum=0}
BEGIN		{tapnum=0}
BEGIN		{wirenum=0}
BEGIN		{i_valnum=0}
BEGIN {old_obj = "x"; old_att ="x"}
BEGIN \
{
wirenum=0;

line_points=0;
line=0;
}

# FSM variables
BEGIN \
{
statenum=0;
actnum=0;
decnum=0;
fsmlabelnum=0;
fsmnum=0;
def_name="";

fsm_clk_name="CLK";
fsm_clk="CLK'event and CLK='1'";
fsm_sync_reset="false";
fsm_rst_name="RST";
fsm_rst="RST = '1'";
fsm_y0="";


fsm_object="none";
register_wire=0;
reg_line=0;
reg_black_text=0;
reg_blue_text=0;
reg_red_text=0;
y0line=0;
bt_line=0;
register_y0=0;
chk_y0=0;
}

# Truth Table Variables
BEGIN {
ttabnum=0;
} 

# Flow Diagram Variables
BEGIN {
waitnum=0;
joinnum=0;
}



function abs(num, \
	     result)
 {
   result=num;
   if (num<0) result=num*-1; 
   return result;
 }


#
# identifying the entity definition
#
/^sym/			{object="entity";}
/^\t.###ENT/		{attribute="entity";}
/^attr\(\"entity=\"/	{if (object=="entity" && attribute=="entity")
		    	{entity=substr($2, 2, length($2)-3);
		     	 attribute="arch";}
			}
/^attr\(\"arch=\"/	{if (object=="entity" && attribute=="arch")
		    	{arch=substr($2, 2, length($2)-3);
		     	 attribute="none";}
			}
/^\t\"###INP\"/	{if (object=="entity")
		     {
			attribute="port";
			port_dir[portnum]="in";
		     }
		}
/^\t\"input=/	{if (object=="entity" && attribute=="port")
		     {
			port_name[portnum]=substr($1, 8, length($1)-11);
			attribute="port_type";
		     }
		}
/^\t\"###OUP\"/	{if (object=="entity")
		     {
			attribute="port";
			port_dir[portnum]="out";
		     }
		}
/^\t\"output=/	{if (object=="entity" && attribute=="port")
		     {
			port_name[portnum]=substr($1, 9, length($1)-12);
			attribute="port_type";
		     }
		}
/^\t\"###BID\"/	{if (object=="entity")
		     {
			attribute="port";
			port_dir[portnum]="inout";
		     }
		}
/^\t\"inout=/	{if (object=="entity" && attribute=="port")
		     {
			port_name[portnum]=substr($1, 8, length($1)-11);
			attribute="port_type";
		     }
		}
/^\t\"type=/	{if (object=="entity" && attribute=="port_type")
		     {
			port_id[portnum]= -1;
			port_type[portnum]= \
				substr(($0), 8, length($0)-11);
			#port_type[portnum]=substr($1, 7, length($1)-10);
			#if ((tolower($2) == "to")||(tolower($2) == "downto"))
			#  {
			#   port_type[portnum]=substr(($1 " " $2 " " $3), 7, \
			#   length($1 $2 $3)-8);
			#  };
			attribute="pin_pos";
		     }
		}
#
# extracting data about the position of entity pins
#
#/^poly\(\'black\'/  \ in order to allow red pins also:
#*/^poly\(\'/  \
#*		{if (object=="entity" && attribute=="pin_pos")
#*	   	  {
#*			attribute="area";
#*			FS=",";
#*		  }
#*		}
/^poly\('/  \
		{if (object=="entity" && attribute=="pin_pos")
	   	  {
			attribute="area";
			FS=",";
		  }
		}
/]/ \
		{if (object=="entity" && attribute=="area")
		     {
			last_tuppel="false";
		   	for (n=0; last_tuppel=="false"; n++)
			  {
				epx[portnum, n]=$(2*n+1);
				epx[portnum, n]++;
				epx[portnum, n]--;
				epy[portnum, n]=$(2*n+2);
				if (substr($(2*n+2),length($(2*n+2)), \
				    length($(2*n+2))+2) == "]") \
					last_tuppel="true";
			  };
			epoints[portnum]=n;
			py[portnum, n-1]= \
			 substr(epy[portnum, n-1], 1, \
			 length(epy[portnum, n-1])-1);
			# ensure type is integer
		   	for (n2=0; n2<n; n2++)
			  {
			   epy[portnum, n2]++;
			   epy[portnum, n2]--;
			  };
			portnum++ ;
			FS=" ";
			attribute="none";
		     }
		}


# entity's generic port
/^\t\"###GEN\"/	{if (object=="entity")
		     {
			attribute="gport";
		     }
		}
/^\t\"generic=/	{if (object=="entity" && attribute=="gport")
		     {
			gport_name[gportnum]=substr($1, 10, length($1)-13);
			attribute="gport_type";
		     }
		}
/^\t\"type=/	{if (object=="entity" && attribute=="gport_type")
		     {
			gport_type[gportnum]= \
				substr(($0), 8, length($0)-11);
			#gport_type[gportnum]=substr($1, 7, length($1)-10);
			#if ((tolower($2) == "to")||(tolower($2) == "downto"))
			#  {
			#   gport_type[gportnum]=substr(($1 " " $2 " " $3), 7, \
			#   length($1 $2 $3)-8);
			#  };
			attribute="gport_init";
		     }
		}
/^\t\"init=/	{if (object=="entity" && attribute=="gport_init")
		     {
			gport_init[gportnum]=substr($1, 7, length($1)-9);
			# -> substitute `\"' with `"'
			gsub("\\" "\\", "", gport_init[gportnum]);
			gportnum++;
			attribute="none";
		     }
		}

#
# identifying the components
#
/^icon/		{object="component";
		     attribute="none";
		     pinnum[comnum]=0;
		     gpinnum[comnum]=0;
		     comnum++;  
		}
/^\t\"###ENT\"/	{if (object=="component") attribute="component"}
/^\t\"entity=/	{if (object=="component" && attribute=="component")
		    {
		     value=substr($1, 9, length($1)-13);
		     new_component[comnum-1]="true";
		     for (i in component)
		       if (value==component[i]) new_component[comnum-1]="false";
		     component[comnum-1]=value;
		    }
		}
/^\t\"###INP\"/	{if (object=="component")
		     {
			attribute="pin";
			pin_dir[comnum-1, pinnum[comnum-1]]="in";
		     }
		}
/^\t\"###OUP\"/	{if (object=="component")
		     {
			attribute="pin";
			pin_dir[comnum-1, pinnum[comnum-1]]="out";
		     }
		}
/^\t\"###BID\"/	{if (object=="component")
		     {
			attribute="pin";
			pin_dir[comnum-1, pinnum[comnum-1]]="inout";
		     }
		}
/^\t\"###GEN\"/	{if (object=="component")
		     {
			attribute="gpin";
		     }
		}
/^\t\"input=/	{if (object=="component" && attribute=="pin")
		     {
			pin_name[comnum-1, pinnum[comnum-1]]= \
			substr($1, 8, length($1)-11);
			attribute="pin_type";
		     }
		}
/^\t\"output=/	{if (object=="component" && attribute=="pin")
		     {
			pin_name[comnum-1, pinnum[comnum-1]]= \
			substr($1, 9, length($1)-12);
			attribute="pin_type";
		     }
		}
/^\t\"inout=/	{if (object=="component" && attribute=="pin")
		     {
			pin_name[comnum-1, pinnum[comnum-1]]= \
			substr($1, 8, length($1)-11);
			attribute="pin_type";
		     }
		}
/^\t\"generic=/	{if (object=="component" && attribute=="gpin")
		     {
			gpin_name[comnum-1, gpinnum[comnum-1]]= \
			substr($1, 10, length($1)-13);
			attribute="gpin_type";
		     }
		}
/^\t\"type=/	{if (object=="component" && attribute=="pin_type")
		     {
			pin_id[comnum-1, pinnum[comnum-1]]= -1;
			pin_type[comnum-1, pinnum[comnum-1]]= \
				substr(($0), 8, length($0)-11);
			#pin_type[comnum-1, pinnum[comnum-1]]= \
			#substr($1, 7, length($1)-10);
			#if ((tolower($2) == "to")||(tolower($2) == "downto"))
			#  {
			#   pin_type[comnum-1, pinnum[comnum-1]]= \
			#   substr(($1 " " $2 " " $3), 7, \
			#   length($1 $2 $3)-8);
			#  };
			attribute="pin_pos";
		     } \
		}

#
# extracting data about the position of component pins
#
#/^poly\('black'/  \ in order to allow red pins also:
/^poly\('/  \
		{if (object=="component" && attribute=="pin_pos")
	   	  {
			attribute="area";
			FS=",";
		  }
		}
/]/ \
		{if (object=="component" && attribute=="area")
		     {
			last_tuppel="false";
		   	for (n=0; last_tuppel=="false"; n++)
			  {
				ipx[comnum-1, pinnum[comnum-1], n]=$(2*n+1);
				ipx[comnum-1, pinnum[comnum-1], n]++;
				ipx[comnum-1, pinnum[comnum-1], n]--;
				ipy[comnum-1, pinnum[comnum-1], n]=$(2*n+2);
				if (substr($(2*n+2),length($(2*n+2)), \
				    length($(2*n+2))+2) == "]") \
					last_tuppel="true";
			  };
			ipoints[comnum-1, pinnum[comnum-1]]=n;
			ipy[comnum-1, pinnum[comnum-1], n-1]= \
			 substr(ipy[comnum-1, pinnum[comnum-1], n-1], 1, \
			 length(ipy[comnum-1, pinnum[comnum-1], n-1])-1);
			# ensure type is integer
		   	for (n2=0; n2<n; n2++)
			  {
			   ipy[comnum-1, pinnum[comnum-1], n2]++;
			   ipy[comnum-1, pinnum[comnum-1], n2]--;
			  };
			pinnum[comnum-1]++ ;
			FS=" ";
			attribute="none";
		     }
		}
# generic pins
/^\t\"type=/	{if (object=="component" && attribute=="gpin_type")
		     {
			gpin_type[comnum-1, gpinnum[comnum-1]]= \
				substr(($0), 8, length($0)-11);
			#substr($1, 7, length($1)-10);
			#if ((tolower($2) == "to")||(tolower($2) == "downto"))
			#  {
			#   gpin_type[gpinnum]=substr(($1 " " $2 " " $3), 7, \
			#   length($1 $2 $3)-8);
			#  }; \
			FS=",";
			attribute="gpin_area";
		     }
		}
/],/ \
		{# record position of generic pin
		 if (object=="component" && attribute=="gpin_area")
		     {
		   	for (n=0; n<6; n++)
			  {
			   gpx[comnum-1, gpinnum[comnum-1], n]=$(2*n+1)+1-1;
			   gpy[comnum-1, gpinnum[comnum-1], n]=$(2*n+2)+1-1;
			  };
			FS=" ";
			attribute="gpin_init";
		     }
		}

/^\t\"init=/	{if (object=="component" && attribute=="gpin_init")
		     {
			gpin_init[comnum-1, gpinnum[comnum-1]]= \
			substr($1, 7, length($1)-9);
			# -> substitute `\"' with `"'
			gsub("\\" "\\", "", gpin_init[comnum-1, \
					    gpinnum[comnum-1]]);
			gpinnum[comnum-1]++ ;
			attribute="none";
		     }
		}



#
# building a list of labels
#
/^\t\"###label\"/	{
				object="label";
				attribute="wire";
			}
/^poly\('blue'/	{if (object=="label" && attribute=="wire")
		   	  {
				attribute="line1";
				FS=",";
			  }
			}
/]/			{if (object=="label" && attribute=="line1")
		   	  {
				l1x0[labelnum]=$1;
				l1x0[labelnum]++;l1x0[labelnum]--;
				l1y0[labelnum]=$2;
				l1y0[labelnum]++;l1y0[labelnum]--;
				l1x1[labelnum]=$3;
				l1x1[labelnum]++;l1x1[labelnum]--;
				l1y1[labelnum]=substr($4, 1, length($4)-1);
				l1y1[labelnum]++;l1y1[labelnum]--;
				attribute="wire2";
				FS=" ";
			  }
			}
/^poly\('blue'/	{if (object=="label" && attribute=="wire2")
		   	  {
				attribute="line2";
				FS=",";
			  }
			}
/]/			{if (object=="label" && attribute=="line2")
		   	  {
				l2x0[labelnum]=$1;
				l2x0[labelnum]++;l2x0[labelnum]--;
				l2y0[labelnum]=$2;
				l2y0[labelnum]++;l2y0[labelnum]--;
				l2x1[labelnum]=$3;
				l2x1[labelnum]++;l2x1[labelnum]--;
				l2y1[labelnum]=substr($4, 1, length($4)-1);
				l2y1[labelnum]++;l2y1[labelnum]--;
				attribute="attribute";
				FS=" ";
			  }
			}
/^\t\"type=/	{if (object=="label" && attribute=="attribute")
		     {
			labeltype[labelnum]= substr(($0), 8, length($0)-11);
			attribute="attribute";
		     }
		}
/^\t\"name=/	{if (object=="label" && attribute=="attribute")
		     {
			labelname[labelnum]= substr($1, 7, length($1)-9);
			label_id[labelnum]= -1;
			labelnum++;
			object="none";
		     }
		}



#
# building a list of tabs
#
/^\t\"###bus_in\"/	{
				object="tap";
				attribute="wire";
				tap_dir[tapnum]= "bus_in";
			}
/^\t\"###bus_out\"/	{
				object="tap";
				attribute="wire";
				tap_dir[tapnum]= "bus_out";
			}
/^\t\"###bus_inout\"/	{
				object="tap";
				attribute="wire";
				tap_dir[tapnum]= "bus_inout";
			}
/^polygon\('blue'/	{if (object=="tap" && attribute=="wire")
		   	  {
				attribute="position";
				FS=",";
			  }
			}
/]/			{if (object=="tap" && attribute=="position")
		   	  {
				bus_x0[tapnum]=$1;
				bus_x0[tapnum]++;bus_x0[tapnum]--;
				bus_y0[tapnum]=$2;
				bus_y0[tapnum]++;bus_y0[tapnum]--;
				bus_x1[tapnum]=$3;
				bus_x1[tapnum]++;bus_x1[tapnum]--;
				bus_y1[tapnum]=$4;
				bus_y1[tapnum]++;bus_y1[tapnum]--;

				tap_x0[tapnum]=$5;
				tap_x0[tapnum]++;tap_x0[tapnum]--;
				tap_y0[tapnum]=$6;
				tap_y0[tapnum]++;tap_y0[tapnum]--;
				tap_x1[tapnum]=$7;
				tap_x1[tapnum]++;tap_x1[tapnum]--;
				tap_y1[tapnum]=$8;
				tap_y1[tapnum]++;tap_y1[tapnum]--;

				attribute="range";
				FS=" ";
			  }
			}
/^\t\"/	{if (object=="tap" && attribute=="range")
		     {
			tap_range[tapnum]= substr(($0), 3, length($0)-6);
			bus_id[tapnum]= -1;
			tap_id[tapnum]= -1;
			tapnum++;
			attribute="arrow";
		     }
		}
# absorb the arrows
/^poly\('blue'/	{if (object=="tap" && attribute=="arrow")
			   attribute="arrow_end";
			}
/]/			{if (object=="tap" && attribute=="arrow_end")
  			   object="none";
			}


#
# building a list of init values
#
/^\t\"###VAL\"/	{
		  object="i_val";
		}
/^\t\"\(/	{if (object=="i_val")
		     {
			i_val[i_valnum]= substr(($0), 4, length($0)-8);
		        FS=",";
		     }
		}
/],/	{if (object=="i_val")
	     {
		i_val_x[i_valnum]=$3 +1-1;
		i_val_y[i_valnum]=$4 +1-1;
		i_valnum++;
		object="none";
		FS=" ";
	     }
	}




#
# collecting data about wiring
#
/^poly\('blue'/	{if (object!="label" && object!="tap" && \
			     object!="i_val")
		   	  {
				object="wire";
				attribute="position";
				split($0, def_string, ",");
				points[wirenum]=def_string[2];
				FS=",";
		   	  }
			}
/]/			{if (object=="wire" && attribute=="position")
		   	  {
			   last_tuppel="false";
		   	   for (n=0; n<points[wirenum]; n++)
			     {
				wx[wirenum, n]=$(2*n+1)+1-1;
				wy[wirenum, n]=$(2*n+2);
			     };
			   wy[wirenum, n-1]= \
			      substr(wy[wirenum, n-1], 1, \
				     length(wy[wirenum, n-1])-1);
			   # ensure all points become integers
		   	   for (n=0; n<=points[wirenum]; n++)
			     {
				wx[wirenum, n]++; wx[wirenum, n]--;
				wy[wirenum, n]++; wy[wirenum, n]--;
			     };
			   wire_name[wirenum]="#none#";
			   wire_type[wirenum]="#none#";
			   wire_is_subnet[wirenum]="false";
			   id[wirenum]=wirenum;
			   object="none";
			   wirenum++; 
			   FS=" ";
			  }
			}



#*{if((object != old_obj)||(attribute!=old_att)){
#*  print "object = " object;
#*  print "attribute = " attribute;
#*  old_obj = object;
#*  old_att = attribute
#* }
#*}


# parse FSM specific objects
# and truthtable objects
#---------------------------

# check object label
/^\t\"###/ \
  {
   split($1,parts,"\"");
   fsm_object=tolower(substr(parts[2],4,length(parts[2])-3));
   #*print "fsm_object = " fsm_object;
  }

# a red line indicates a label-object
/^poly\('red'/  \
  {
   fsm_object="label";
  }


# register lines
/]/ \
  {
   if(reg_line)
     {
      for(i=1;i<=4;i++)
        {
          lp[line_points]=$(i);
	  line_points++;
        }
      lp[line_points-1]=substr($4,1,length($4)-1);
      reg_line=0;
      FS=" ";
     }
  }
/^poly\('/  \
  {
   reg_line=1;
   FS=",";
  }


# register black text
/^\t\"/ \
  {
   if(reg_black_text)
     {
      black_text[line]=substr($0,3,length($0)-4);
      line++;
     }
  }
/\"]\)/   \
  {
   if(reg_black_text) reg_black_text=0;
  }
/^text\('black'/  \
  {
   reg_black_text=1;
  }



# register blue text
/^\t\"/ \
  {
   if(reg_blue_text)
     {
      blue_text[bt_line]=substr($0,3,length($0)-4);
      gsub("\\" "\\", "", blue_text[bt_line]);
      bt_line++;
     }
  }
/\"]\)/   \
  {
   if(reg_blue_text) reg_blue_text=0;
  }
/^text\('blue'/  \
  {
   reg_blue_text=1;
  }


# register Y0
/^\t\"/ \
  {
   if(chk_y0 && match($0,"Y0:")) register_y0=1;
   if(register_y0)
     {
      y0[y0line]=substr($0,3,length($0)-4);
      gsub("Y0:","",y0[y0line]);
      y0line++;
     }
  }
/\"]\)/   \
  {
   chk_y0=0;
   register_y0=0;
  }
/^text\('blue'/  \
  {
   chk_y0=1;
  }


# register red text
/^\t\"/ \
  {
   if(reg_red_text)
     {
      red_text[line]=substr($0,3,length($0)-4);
      # eliminate spare `"' when on last line
      if(substr($0,length($0))==")") red_text[line]=substr($0,3,length($0)-5);
      if(substr($0,length($0)-1,2)=="),")
         red_text[line]=substr($0,3,length($0)-6);
      gsub("\\" "\\", "", red_text[line])
      line++;
     }
  }
/\"]\)/   \
  {
   if(reg_red_text) reg_red_text=0;
  }
/^text\('red'/  \
  {
   reg_red_text=1;
  }


# register diagonal points of a box
/^box\('/ \
  {
   split($0,parts,",");
   bp_x[0]=parts[2];
   bp_y[0]=parts[3];
   bp_x[1]=parts[4];
   bp_y[1]=parts[5];
   #print "#### " bp_x[0] "." bp_y[0] "/" bp_x[1] "." bp_y[1]
  }


# end of a group, assemble FSM object
/^])\./ \
  {
   if(lp[1] < lp[5]) # first line indicates input
     {
      in_x1=lp[0];
      in_y1=lp[1];
      in_x2=lp[2];
      in_y2=lp[3];
      out_x1=lp[4];
      out_y1=lp[5];
      out_x2=lp[6];
      out_y2=lp[7];
     }
   if(lp[1] > lp[5]) # second line indicates input
     {
      in_x1=lp[4];
      in_y1=lp[5];
      in_x2=lp[6];
      in_y2=lp[7];
      out_x1=lp[0];
      out_y1=lp[1];
      out_x2=lp[2];
      out_y2=lp[3];
     }

   for(l=0;l<line;l++) gsub("\\" "\\", "", black_text[l]);
   black_text[line-1] = substr(black_text[line-1],1,     \
                        length(black_text[line-1])-2);
   blue_text[bt_line-1] = substr(blue_text[bt_line-1],1,     \
                        length(blue_text[bt_line-1])-2);

   #*print "";
   #*for(l=0;l<line;l++) print "###" red_text[l] "#";


   if(fsm_object=="state")
     {
      state_in_x1[statenum]=in_x1;
      state_in_y1[statenum]=in_y1;
      state_in_x2[statenum]=in_x2;
      state_in_y2[statenum]=in_y2;
      state_out_x1[statenum]=out_x1;
      state_out_y1[statenum]=out_y1;
      state_out_x2[statenum]=out_x2;
      state_out_y2[statenum]=out_y2;
      state_name[statenum]=black_text[0];
      state_rline_x0[statenum]=bp_x[1];
      state_rline_y0[statenum]=bp_y[0];
      state_rline_x1[statenum]=bp_x[1];
      state_rline_y1[statenum]=bp_y[1];
      statenum++;
     }

   if(fsm_object=="action")
     {
      act_in_x1[actnum]=in_x1;
      act_in_y1[actnum]=in_y1;
      act_in_x2[actnum]=in_x2;
      act_in_y2[actnum]=in_y2;
      act_out_x1[actnum]=out_x1;
      act_out_y1[actnum]=out_y1;
      act_out_x2[actnum]=out_x2;
      act_out_y2[actnum]=out_y2;
      #*for(l=0;l<line;l++) act_vhdl[actnum, l]=black_text[l];
      #*act_vhdl_lines[actnum]=line;
      for(l=0;l<line;l++) act_vhdl[actnum]=act_vhdl[actnum] black_text[l] " ";
      gsub("; ",";",act_vhdl[actnum]);
      gsub(";",";\n",act_vhdl[actnum]);
      gsub("\n$","",act_vhdl[actnum]);
      actnum++;
     }

   if(fsm_object=="decision")
     {
      dec_in_x1[decnum]=in_x1;
      dec_in_y1[decnum]=in_y1;
      dec_in_x2[decnum]=in_x2;
      dec_in_y2[decnum]=in_y2;
      dec_out_x1[decnum]=out_x1;
      dec_out_y1[decnum]=out_y1;
      dec_out_x2[decnum]=out_x2;
      dec_out_y2[decnum]=out_y2;
      for(l=0;l<line;l++) dec_expr[decnum]=dec_expr[decnum] black_text[l] " ";
      decnum++;
     }

   # info for definition object
   for(l=0;l<bt_line;l++)
     {
      if(substr(blue_text[l],1,4)=="FSM=")
        {fsm_string=substr(blue_text[l],5,length(blue_text[l])-6);}
      if(match(blue_text[l],"Clock Signal:"))
        {clk_name=substr(blue_text[l],14,length(blue_text[l])-15);}
       if(match(blue_text[l],"Clock Condition:"))
        {clk_string=substr(blue_text[l],17,length(blue_text[l])-18);}
     if(match(blue_text[l],"Reset Signal:"))
        {rst_name=blue_text[l]; gsub("Reset Signal:","",rst_name);}
      if(match(blue_text[l],"Reset Condition:"))
        {rst_string=substr(blue_text[l],17,length(blue_text[l])-18);}
     if(match(blue_text[l],"sync_reset"))
        {
         sync_reset="false";
         if(match(tolower(blue_text[l]),"true")) sync_reset="true";
        }
     }

   if(fsm_object=="def")
     {
      def_name=fsm_string;
      fsm_clk_name=clk_name;
      fsm_clk=clk_string;
      fsm_rst_name=rst_name;
      fsm_rst=rst_string;
      fsm_sync_reset=sync_reset;
      for(l=0;l<y0line;l++) gsub("\\" "\\", "", y0[l]);
      y0[y0line-1] = substr(y0[y0line-1],1,length(y0[y0line-1])-2);
      for(l=0;l<y0line;l++) fsm_y0=fsm_y0 "\t" y0[l] "\n";
      y0line=0;
     }

   if(fsm_object=="label")
     {
      if(   ((lp[0]-lp[2])**2 + (lp[1]-lp[3])**2)   \
          > ((lp[4]-lp[6])**2 + (lp[5]-lp[7])**2) )
        {
	 # first line is the wire
         lbl_x1[fsmlabelnum] = lp[0]; lbl_y1[fsmlabelnum] = lp[1];
         lbl_x2[fsmlabelnum] = lp[2]; lbl_y2[fsmlabelnum] = lp[3]
        }
       else
        {
	 # second line is the wire
         lbl_x1[fsmlabelnum] = lp[4]; lbl_y1[fsmlabelnum] = lp[5];
         lbl_x2[fsmlabelnum] = lp[6]; lbl_y2[fsmlabelnum] = lp[7];
        }
 
      #register label expression
      for(l=0;l<line;l++) 
	  lbl_expr[fsmlabelnum]=lbl_expr[fsmlabelnum] red_text[l] " ";
      fsmlabelnum++;
      # add the blue lines as wires
      # - upper line
      #wx[wirenum, 0]=in_x1;
      #wy[wirenum, 0]=in_y1;
      #wx[wirenum, 1]=in_x2;
      #wy[wirenum, 1]=in_y2;
      #points[wirenum]=2;
      #wire_name[wirenum]="#none#";
      #wire_type[wirenum]="#none#";
      #wire_is_subnet[wirenum]="false";
      #id[wirenum]=wirenum;
      #wirenum++; 
      # - lower line
      #wx[wirenum, 0]=out_x1;
      #wy[wirenum, 0]=out_y1;
      #wx[wirenum, 1]=out_x2;
      #wy[wirenum, 1]=out_y2;
      #points[wirenum]=2;
      #wire_name[wirenum]="#none#";
      #wire_type[wirenum]="#none#";
      #wire_is_subnet[wirenum]="false";
      #id[wirenum]=wirenum;
      #wirenum++; 

     }

   if(fsm_object=="start")
     {
      start_x[fsmnum]=lp[0];
      start_y[fsmnum]=lp[1];
      if(lp[2]<lp[0]) {
        start_x[fsmnum]=lp[2];
        start_y[fsmnum]=lp[3];
       }
      gsub("FSM=", "",black_text[0])
      fsm_name[fsmnum]=black_text[0];
      fsmnum++;
     }


   # additional objects for flow diagrams

   if(fsm_object=="wait")
     {
      wait_in_x1[waitnum]=in_x1;
      wait_in_y1[waitnum]=in_y1;
      wait_in_x2[waitnum]=in_x2;
      wait_in_y2[waitnum]=in_y2;
      wait_out_x1[waitnum]=out_x1;
      wait_out_y1[waitnum]=out_y1;
      wait_out_x2[waitnum]=out_x2;
      wait_out_y2[waitnum]=out_y2;
      wait_vhdl[waitnum]="";
      for(l=0;l<line;l++) 
         wait_vhdl[waitnum]=wait_vhdl[waitnum] black_text[l] " ";
      waitnum++;
     }



   if(fsm_object=="join")
     {
      join_in1_x1[joinnum]=in_x1;
      join_in1_y1[joinnum]=in_y1;
      join_in1_x2[joinnum]=in_x2;
      join_in1_y2[joinnum]=in_y2;
      join_in2_x1[joinnum]=side_x1;
      join_in2_y1[joinnum]=side_y1;
      join_in2_x2[joinnum]=side_x2;
      join_in2_y2[joinnum]=side_y2;
      join_out_x1[joinnum]=out_x1;
      join_out_y1[joinnum]=out_y1;
      join_out_x2[joinnum]=out_x2;
      join_out_y2[joinnum]=out_y2;
      joinnum++;
     }


   fsm_object="none";
   line_points=0;
   line=0;
   bt_line=0;
 }



# Store Truthtable information

/\"])\.$/ {   # end of a text block

   if(match(fsm_object, "truthtable")) {
	for(tl=0;tl<bt_line;tl++) truthline[ttabnum,tl] = blue_text[tl];
	gsub("\"\]","",truthline[ttabnum,bt_line-1]);
	ttablines[ttabnum]=bt_line;
	ttabnum++;

	fsm_object="none";
	bt_line=0;
     }

  }



#
#  Processing the data
#




# grouping wires to nets

# identifying whether a point
# is on a horizantal or vertical segment
#
function point_on_seg(ptx, pty, wire, pt, \
		      y0, y1, on_seg)
 {
	on_seg="false";
	# point on vertical segment
	if (wy[wire,s]>wy[wire,s+1])
  	   {y0=wy[wire,pt+1]; y1=wy[wire,pt]}
	 else
 	   {y0=wy[wire,pt]; y1=wy[wire,pt+1]};
	if (ptx==wx[wire,pt] && wx[wire,pt]==wx[wire,pt+1] && \
	    pty>=y0 && pty<=y1)
	    on_seg="true";

	# point on horizontal segment
	if (wx[wire,pt]>wx[wire,pt+1])
		{x0=wx[wire,pt+1]; x1=wx[wire,pt]}
	  else
		{x0=wx[wire,pt]; x1=wx[wire,pt+1]};
	if (pty==wy[wire,pt] && wy[wire,pt]==wy[wire,pt+1] && \
	    ptx>=x0 && ptx<=x1)
	    on_seg="true";

	return on_seg;
 }



# identifying whether a line 
# is on a horizontal or vertical segment
#
function line_on_seg(lx0, ly0, lx1, ly1, wire, pt, \
		      a, b, w0, w1, on_seg)
 {

  	on_seg="false";

	# overlapping vertical segments
	if (lx0==lx1 && wx[wire,pt]==wx[wire,pt+1] && \
	    lx0==wx[wire,pt])
	  {
 	    if (ly0>ly1)
  		{a=ly1; b=ly0}
  	       else
  		{a=ly0; b=ly1};
  	    if (wy[wire,pt]>wy[wire,pt+1])
  		{w0=wy[wire,pt+1]; w1=wy[wire,pt]}
  	       else
  		{w0=wy[wire,pt]; w1=wy[wire,pt+1];}

	    if ( (w0>=a && w0<=b) || \
  	         (w1>=a && w1<=b) || \
	         (a>=w0 && a<=w1) )
  	         on_seg="true";
 	   };

  	# overlapping horizontal segments
  	if (ly0==ly1 && wy[wire,pt]==wy[wire,pt+1] && \
  	    ly0==wy[wire,pt])
  	  {
  	    if (lx0>lx1)
  		{a=lx1; b=lx0}
  	       else
  		{a=lx0; b=lx1};
  	    if (wx[wire,pt]>wx[wire,pt+1])
  		{w0=wx[wire,pt+1]; w1=wx[wire,pt]}
  	       else
  		{w0=wx[wire,pt]; w1=wx[wire,pt+1];}

	    if ( (w0>=a && w0<=b) || \
  	         (w1>=a && w1<=b) || \
	         (a>=w0 && a<=w1) )
  	         on_seg="true";
  	  };

	return on_seg;

}



# identifying whether a point
# is on a horizontal or vertical line

function point_on_line(ptx, pty, lx0, ly0, lx1, ly1, \
		      y0, y1, x0, x1, on_line)
 {
	on_line="false";
	#print "###";
	#print "lx0 " lx0 "lx1 " lx1 "ptx" ptx;

	# point on vertical line
	if (lx0==lx1 && ptx==lx0)
	  {
	   y0=ly0; y1=ly1;
	   if (ly0>ly1) {y0=ly1; y1=ly0}
           if(pty>=y0 && pty<=y1) on_line="true";
	  }

	# point on horizontal line
	if (ly0==ly1 && pty==lx0)
	  {
	   x0=lx0; x1=lx1;
	   if (lx0>lx1) {x0=lx1; x1=lx0}
           if(ptx>=x0 && ptx<=x1) on_line="true";
	  }

	return on_line;
 }


# splitting wires
# (a red label is supposed to cut any wires which it overlaps)

END {
# go through all wires
if(0) # switch off for now
for(w=0; w<wirenum-1; w++) {
   # for all wire segments of w1
   for(s=0; s<points[w]-2; s++) {
      # for all FSM labels
      for(l=0; l<fsmlabelnum; l++) {
	# connected to the first label point only
	# make sure its not already connected to an endpoint
        if(point_on_seg(lbl_x1[l], lbl_y1[l], w, s) == "true" && \
	   point_on_seg(lbl_x2[l], lbl_y2[l], w, s) == "false" && \
	   (lbl_x1[l]!=wx[w,s] || lbl_y1[l]!=wy[w,s]) && \
	   (lbl_x1[l]!=wx[w,s+1] || lbl_y1[l]!=wy[w,s+1]) ) {
          print "### label " l " connected to wire " w;

	  # cut original wire

	  # add rest of wire as a new wire
	  points[wirenum]=points[w]-s;
	  for(t=1; t<points[wirenum]-1; t++)
           {
	    wx[wirenum,t]=wx[w,s+t+1];
	    wy[wirenum,t]=wy[w,s+t+1];
	   }
	  wire_name[wirenum]="#none#";
	  wire_type[wirenum]="#none#";
	  wire_is_subnet[wirenum]="false";
	  id[wirenum]=wirenum;
	  wirenum++;

	  # change connecting segment so that it goes 
	  # directly to the label point
	  wx[w,s]=lbl_x1[l];
	  wy[w,s]=lbl_y1[l];
	  points[w]=s+2;
	     for(t=1; t<points[wirenum]-1; t++)
              {
	       wx[wirenum,t]=wx[w,s+t+1];
	       wy[wirenum,t]=wy[w,s+t+1];
	      }
	     wire_name[wirenum]="#none#";
	     wire_type[wirenum]="#none#";
	     wire_is_subnet[wirenum]="false";
	     id[wirenum]=wirenum;
	     wirenum++;

	  # add remainder of wire as a new wire

	  if(lbl_x1[l]==lbl_x1[l]) {
	    x_end=lbl_x1[l];
	    x_start=lbl_x1[l]}
	   else {
	    y_end=lbl_y1[l];
	    y_start=lbl_y1[l]};
	  #if(wx[w,s]==wx[w,s+1] && 

	  # cut original wire
	  points[w]=s+1;
	  #wx[w,s]=;
	  # add remainder of wire as a new wire
	  if(s<points[w]-2)
            {
	     points[wirenum]=points[w]-s;
	     # other end of label is first point of remaining wire
	     # one contact only:

	     #wx[wirenum,t]=;
	     #wy[wirenum,t]=;
	     # double contact:

	     for(t=1; t<points[wirenum]-1; t++)
              {
	       wx[wirenum,t]=wx[w,s+t+1];
	       wy[wirenum,t]=wy[w,s+t+1];
	      }
	     wire_name[wirenum]="#none#";
	     wire_type[wirenum]="#none#";
	     wire_is_subnet[wirenum]="false";
	     id[wirenum]=wirenum;
	     wirenum++;
            }
         }

      }
     }
  }
}




END \
{
   # for all wires w1
   # (last wire will not be compared since there 
   #  will be nothing left to compare it with)
   for (w1=0; w1<wirenum-1; w1++)
    {
       # for all wire segments of w1
       for (s=0; s<points[w1]-1; s++)
        {
 	# for all other wires w2
         for (w2=w1+1; w2<=wirenum-1; w2++)
 	  {
 	   wires_connected="false";
  
 	   # matching endpoints 
	   #
	   # -first and first
 	   w1x=wx[w1,0]; w1y=wy[w1,0];
  	   w2x=wx[w2,0]; w2y=wy[w2,0];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";
	   # -first and last
	   w2x=wx[w2,points[w2]-1]; w2y=wy[w2,points[w2]-1];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";
	   # -last and last
	   w1x=wx[w1,s+1]; w1y=wy[w1,s+1];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";
	   # -last and first
  	   w2x=wx[w2,0]; w2y=wy[w2,0];
	   if (w1x==w2x && w1y==w2y) wires_connected="true";

	   # check whether w2's endpoints are on current w1 segment
	   #
	   # first point on segment
	   if (point_on_seg(wx[w2,0], wy[w2,0], w1, s)=="true")
		wires_connected="true";
	   # last point on segment
	   if (point_on_seg(wx[w2,points[w2]-1], \
			    wy[w2,points[w2]-1], w1, s) == "true")
		wires_connected="true";

  
  	   # for all segments of w2
             for (t=0; t<=points[w2]-2; t++)
  	      {

		# check whether w1's endpoints are on current w2 segment
		# first on segment
		if (point_on_seg(wx[w1,0], wy[w1,0], w2, t) == "true")
 		   wires_connected= "true";
		# last on segment
		if (point_on_seg(wx[w1,points[w1]-1], \
		 		 wy[w1,points[w1]-1], w2, t) == "true")
		   wires_connected= "true";
   
 		# overlapping segments
		if (line_on_seg(wx[w2,t],wy[w2,t], wx[w2,t+1],wy[w2,t+1], w1,s) \
		    == "true") wires_connected= "true";
		
  	      }

 	    # group together
 	    if (wires_connected=="true")
	      { 
		replace_id=id[w2];
		for (v=0;v<wirenum;v++) if(id[v]==replace_id) id[v]=id[w1];
	      } 
 	  }
        }
    }	
 }
 
#~ this code is extremely messy, should be split in functions


# identifying nets which are connected to companent instantiation ports
# + assigning signal names and types
# + assign wire IDs to the component ports

END \
{
#for all components
for (c=0; c < comnum; c++)
 {
 #print "###com=" c;
 # for all component pins
 for (p=0; p < pinnum[c]; p++)
   {
     #print "###pin_dir=" pin_dir[c,p];
     #print "###p0=" ipx[c,p,0] "," ipy[c,p,0];
     #print "###p1=" ipx[c,p,1] "," ipy[c,p,1];
     #print "###p2=" ipx[c,p,2] "," ipy[c,p,2];
     #print "###p3=" ipx[c,p,3] "," ipy[c,p,3];
     #print "###p4=" ipx[c,p,4] "," ipy[c,p,4];
     #print "###p5=" ipx[c,p,5] "," ipy[c,p,5];
     # for all wires
     for (w=0; w < wirenum; w++)
      {
	connected="false";
	#print "###wirenum=" w;
	#print "###w0=" wx[w,0] "," wy[w,0];
	#print "###wn=" wx[w,points[w]-1] "," wy[w,points[w]-1];
	if (pin_dir[c,p]=="in" || pin_dir[c,p]=="out")
	  {
	   # wire's endpoints within the rectangular field

	   if (wx[w,0]>=ipx[c,p,0] && wx[w,0]<=ipx[c,p,1] && \
	       wy[w,0]>=ipy[c,p,0] && wy[w,0]<=ipy[c,p,4])
		connected="true";
	   if (wx[w,points[w]-1]>=ipx[c,p,0] && wx[w,points[w]-1]<=ipx[c,p,1] && \
	       wy[w,points[w]-1]>=ipy[c,p,0] && wy[w,points[w]-1]<=ipy[c,p,4])
		connected="true";
	   # wire's endpoints within the triangular field
 	   if ( wx[w,0]>=ipx[c,p,1] && wx[w,0]<=ipx[c,p,2] && \
 	        (wy[w,0]-ipy[c,p,1])>=(wx[w,0]-ipx[c,p,1]) && \
 	        (wy[w,0]-ipy[c,p,2])<=(ipx[c,p,2]-wx[w,0]) )
 		connected="true";
 	   if ( wx[w,points[w]-1]>=ipx[c,p,1] && wx[w,points[w]-1]<=ipx[c,p,2] && \
	        (wy[w,points[w]-1]-ipy[c,p,1])>=(wx[w,points[w]-1]-ipx[c,p,1]) && \
 	        (wy[w,points[w]-1]-ipy[c,p,2])<=(ipx[c,p,2]-wx[w,points[w]-1]) )
 		connected="true";
	  }
	if (pin_dir[c,p]=="inout")
	  {
	   # wire's endpoints within the rectangular field
	   if (wx[w,0]>=ipx[c,p,1] && wx[w,0]<=ipx[c,p,2] && \
	       wy[w,0]>=ipy[c,p,1] && wy[w,0]<=ipy[c,p,5])
		connected="true";
	   if (wx[w,points[w]-1]>=ipx[c,p,1] && wx[w,points[w]-1]<=ipx[c,p,2] && \
	       wy[w,points[w]-1]>=ipy[c,p,1] && wy[w,points[w]-1]<=ipy[c,p,5])
		connected="true";
	   # wire's endpoints within the right triangular field
 	   if ( wx[w,0]>=ipx[c,p,2] && wx[w,0]<=ipx[c,p,3] && \
	        (wy[w,0]-ipy[c,p,2])>=(wx[w,0]-ipx[c,p,2]) && \
	        (wy[w,0]-ipy[c,p,3])<=(ipx[c,p,3]-wx[w,0]) )
		connected="true";
	   #print "p2=" ipx[c,p,2] "," ipy[c,p,2];
	   #print "p3=" ipx[c,p,3] "," ipy[c,p,3];
	   #print "wn=" wx[w,points[w]-1] "," wy[w,points[w]-1];
	   #print wx[w,points[w]-1]>=ipx[c,p,2]
	   if ( wx[w,points[w]-1]>=ipx[c,p,2] && wx[w,points[w]-1]<=ipx[c,p,3] && \
	        (wy[w,points[w]-1]-ipy[c,p,2])>=(wx[w,points[w]-1]-ipx[c,p,2]) && \
	        (wy[w,points[w]-1]-ipy[c,p,3])<=(ipx[c,p,3]-wx[w,points[w]-1]) )
		connected="true";
 	   # wire's endpoints within the left triangular field
 	   if ( wx[w,0]>=ipx[c,p,0] && wx[w,0]<=ipx[c,p,1] && \
	       (wy[w,0]-ipy[c,p,0])<=(wx[w,0]-ipx[c,p,0]) && \
	       (wy[w,0]-ipy[c,p,1])>=(ipx[c,p,1]-wx[w,0]) )
	       connected="true";
 	   if ( wx[w,points[w]-1]>=ipx[c,p,0] && wx[w,points[w]-1]<=ipx[c,p,1] && \
 	       (wy[w,points[w]-1]-ipy[c,p,0])<=(wx[w,points[w]-1]-ipx[c,p,0]) && \
 	       (wy[w,points[w]-1]-ipy[c,p,1])>=(ipx[c,p,1]-wx[w,points[w]-1]) )
 	       connected="true";
	  }
	#if (connected!="true") {print "wire " w " not connected to pin";}
	if (connected=="true")
	  {
	    # determining name and type of the wire
	    #
	    # output pins define the wire name
	    if (pin_dir[c,p]=="out" && wire_name[w]="#none#")
	        for (v=0; v<wirenum; v++)  
		   if (id[v]==id[w]) wire_name[v]="U" c "_" pin_name[c,p];
	    # inout defines the name if no output does
	    if (pin_dir[c,p]=="inout" && wire_name[w]="#none#")
	        for (v=0; v<wirenum; v++)  
		   if (id[v]==id[w]) 
			wire_name[v]="U" c "_" pin_name[c,p];
            # linking nets which are connected via a pin
	    #print "pin id" pin_id[c,p];
	    if (pin_id[c,p]>=0) 
	       {
		# two wires are connected via a pin
		replace_id=id[w];
		for (v=0; v<wirenum; v++) 
		  { 
		   if (id[v]==replace_id) 
		     { 
		      id[v]=pin_id[c,p];
		      wire_name[v]=wire_name[pin_id[c,p]];
		     }
		  } 
	       }
	      else
	       {
		# associating pin with wire ID
		pin_id[c,p]=id[w];
	       }
	    # assign the right type to all wires of the net
	    for (v=0; v<wirenum; v++)  
		   if (id[v]==id[w]) wire_type[v]=pin_type[c,p];
	  }
      } 
   }
 }
}



# identifying nets which are connected to symbol ports
# (+assigning signal names and types)



# checks whether point is in a rectangle
# (r1 and r2 are diagonal, rx1 < rx2)

function point_in_rectangle(px,py, rx1,ry1,rx2,ry2, connected)
 {
   connected="false";
   if (px>=rx1 && px<=rx2 && py>=ry1 && py<=ry2)
	connected="true";
   return connected;
 }


# checks whether a point is within the triangular part of a pin

function point_in_pin_triangle(px,py, tx1,ty1,tx2,ty2, connected)
 {
   connected="false";
   if (px>=tx1 && px<=tx2 && (py-ty1)>=(px-tx1) && (py-ty2)<=(tx2-px) )
      connected="true";
   return connected;
 }


# checks whether point is within a generic, input or output pin
# (pin points are in clockwise order, starting lower left rectangle pin)

function point_in_pin(px,py, pi1x,pi1y,pi2x,pi2y,pi3x,pi3y,pi4x,pi4y, \
                      connected)
 {
   connected = point_in_rectangle(px,py, pi1x,pi1y, pi4x,pi4y);
   if (connected=="false") 
      connected = point_in_pin_triangle(px,py, pi2x,pi2y,pi3x,pi3y);
   return connected;
 }


END \
{
 # for all symbol ports
 for (p=0; p < portnum; p++)
   {
     #print "### port=" p " " port_dir[p];
     # for all wires
     for (w=0; w < wirenum; w++)
      {
	connected="false";
	#print "### wirenr=" w;
	if (port_dir[p]=="in" || port_dir[p]=="out")
	  {
	   # wire's endpoints within the rectangular field
	   #print "###ep0=" epx[p,0] "," epy[p,0];
	   #print "###ep1=" epx[p,1] "," epy[p,1];
	   #print "###wn=" wx[w,points[w]-1] "," wy[w,points[w]-1];
	   #print "### connected=" connected;
	   if (wx[w,0]>=epx[p,0] && wx[w,0]<=epx[p,1] && \
	       wy[w,0]>=epy[p,0] && wy[w,0]<=epy[p,4])
		connected="true";
	   #print "### connected=" connected;
	   #print "wx[w,points[w]-1]=" wx[w,points[w]-1];
	   #print "wy[w,points[w]-1]=" wy[w,points[w]-1];
	   #print "epx[p,0]=" epx[p,0];
	   #print "epy[p,0]=" epy[p,0];
	   #print "epx[p,1]=" epx[p,1];
	   #print "epy[p,1]=" epy[p,1];
	   #print "epx[p,4]=" epx[p,4];
	   #print "epy[p,4]=" epy[p,4];
	   if (wx[w,points[w]-1]>=epx[p,0] && wx[w,points[w]-1]<=epx[p,1] && \
	       wy[w,points[w]-1]>=epy[p,0] && wy[w,points[w]-1]<=epy[p,4])
		connected="true";
	   #print "### connected=" connected;
	   # wire's endpoints within the triangular field
 	   if ( wx[w,0]>=epx[p,1] && wx[w,0]<=epx[p,2] && \
 	        (wy[w,0]-epy[p,1])>=(wx[w,0]-epx[p,1]) && \
 	        (wy[w,0]-epy[p,2])<=(epx[p,2]-wx[w,0]) )
 		connected="true";
 	   if ( wx[w,points[w]-1]>=epx[p,1] && wx[w,points[w]-1]<=epx[p,2] && \
	        (wy[w,points[w]-1]-epy[p,1])>=(wx[w,points[w]-1]-epx[p,1]) && \
 	        (wy[w,points[w]-1]-epy[p,2])<=(epx[p,2]-wx[w,points[w]-1]) )
 		connected="true";
	  }
	if (port_dir[p]=="inout")
	  {
	   # wire's endpoints within the rectangular field
	   if (wx[w,0]>=epx[p,1] && wx[w,0]<=epx[p,2] && \
	       wy[w,0]>=epy[p,1] && wy[w,0]<=epy[p,5])
		connected="true";
	   if (wx[w,points[w]-1]>=epx[p,1] && wx[w,points[w]-1]<=epx[p,2] && \
	       wy[w,points[w]-1]>=epy[p,1] && wy[w,points[w]-1]<=epy[p,5])
		connected="true";
	   # wire's endpoints within the right triangular field
 	   if ( wx[w,0]>=epx[p,2] && wx[w,0]<=epx[p,3] && \
	        (wy[w,0]-epy[p,2])>=(wx[w,0]-epx[p,2]) && \
	        (wy[w,0]-epy[p,3])<=(epx[p,3]-wx[w,0]) )
		connected="true";
	   #print "p2=" epx[p,2] "," epy[p,2];
	   #print "p3=" epx[p,3] "," epy[p,3];
	   #print "wn=" wx[w,points[w]-1] "," wy[w,points[w]-1];
	   #print wx[w,points[w]-1]>=epx[p,2]
	   if ( wx[w,points[w]-1]>=epx[p,2] && wx[w,points[w]-1]<=epx[p,3] && \
	        (wy[w,points[w]-1]-epy[p,2])>=(wx[w,points[w]-1]-epx[p,2]) && \
	        (wy[w,points[w]-1]-epy[p,3])<=(epx[p,3]-wx[w,points[w]-1]) )
		connected="true";
 	   # wire's endpoints within the left triangular field
 	   if ( wx[w,0]>=epx[p,0] && wx[w,0]<=epx[p,1] && \
	       (wy[w,0]-epy[p,0])<=(wx[w,0]-epx[p,0]) && \
	       (wy[w,0]-epy[p,1])>=(epx[p,1]-wx[w,0]) )
	       connected="true";
 	   if ( wx[w,points[w]-1]>=epx[p,0] && wx[w,points[w]-1]<=epx[p,1] && \
 	       (wy[w,points[w]-1]-epy[p,0])<=(wx[w,points[w]-1]-epx[p,0]) && \
 	       (wy[w,points[w]-1]-epy[p,1])>=(epx[p,1]-wx[w,points[w]-1]) )
 	       connected="true";
	  }
	#if (connected!="true") {print "wire " w " not connected to port";}
	if (connected=="true")
	  {
	    #print "###wire " w " connected to port";
	    #print "###wirenum= " wirenum;
		for (n=0; n < wirenum; n++)
		 { 
			#print "wirenr: " n " name: " wire_name[n];
			#print "id: " id[n]
		 }
	    # determining name and type of the wire 
	    if (port_dir[p]=="out")
	     {
		for (v=0; v<wirenum; v++)  
		   if (id[v]==id[w]) wire_name[v]="i_" tolower(port_name[p]); 
	     }
	    else
	     {
		for (v=0; v<wirenum; v++)  
		   if (id[v]==id[w]) wire_name[v]=port_name[p]; 
	     }
            # linking nets which are connected via a port
	    if (port_id[p]>=0) 
	       {
		# two wires are connected via a port
		replace_id=id[w];
		for (v=0; v<wirenum; v++) 
		  { 
		   if (id[v]==replace_id) 
		     { 
		      id[v]=port_id[p];
		      wire_name[v]=wire_name[port_id[p]];
		     }
		  } 
	       }
	      else
	       {
		# associating port with wire ID
		port_id[p]=id[w];
	       }
	    # assign the right type to all wires of the net
	    for (v=0; v<wirenum; v++)  
		   if (id[v]==id[w]) wire_type[v]=port_type[p];
	  }
      } 
   }
}




# assigning generic names to unnamed wires
END \
{
 for (w=0; w < wirenum; w++)
   if (wire_name[w]=="#none#") wire_name[w]="net_000" id[w];
}


# checking bus type labels
#
#

END \
{
for (n=0;n<labelnum;n++)
  {

	# identifying the "hot" part of the label

	lx0[n]=l1x0[n];lx1[n]=l1x1[n]; 
	ly0[n]=l1y0[n];ly1[n]=l1y1[n]; 
	#print "l1 " l1x0[n] " " l1y0[n] " " l1x1[n] " " l1y1[n];
	#print "l2 " l2x0[n] " " l2y0[n] " " l2x1[n] " " l2y1[n];
	max1=abs(l1y0[n]-l1y1[n]);
	#print "max1=" max1;
	if (abs(l1x0[n]-l1x1[n])>max1) {max1=abs(l1x1[n]-l1x0[n]);}
	#print "max1=" max1;
	if ( (abs(l2x1[n]-l2x0[n])>max1) || (abs(l2y1[n]-l2y0[n])>max1) ) 
		{lx0[n]=l2x0[n];lx1[n]=l2x1[n];
		 ly0[n]=l2y0[n];ly1[n]=l2y1[n]}
	#print "labelnum=" labelnum " wirenum=" wirenum;

	# find the associated wire

	# for all wires w
	for (w=0; w<wirenum; w++)
	  {
	   #print "wire=" w;
	   attached_to_wire="false";
	   # for all wire segments of w1
           for (s=0; s<points[w]-1; s++)
	     {
		# first point on segment
		if (point_on_seg(lx0[n], ly0[n], w, s)=="true")
			attached_to_wire="true";
		#print "attached_to_wire=" attached_to_wire;
		# second point on segment
		if (point_on_seg(lx1[n], ly1[n], w, s) == "true")
			attached_to_wire="true";
		#print "attached_to_wire=" attached_to_wire;
	     }
	   if (attached_to_wire=="true")
	     {
		label_id[n]=id[w];
		# assign the right type to all wires of the net
		for (v=0; v<wirenum; v++)  
		   if (id[v]==id[w]) 
			{ wire_type[v]=labeltype[n]; wire_name[v]=labelname[n]}
	     }
	  }
  }
}



# checking bus taps
# (finding names of bus signals, determing type)

END \
{
for (n=0;n<tapnum;n++)
  {
    for (w=0; w<wirenum; w++)
      {
	#print "wire=" w;
	bus_on_wire="false";
	tap_on_wire="false";
	# for all wire segments of w1
	for (s=0; s<points[w]-1; s++)
	  {
		bus_on_wire="false";
		# first bus point on segment
		if (point_on_seg(bus_x0[n], bus_y0[n], w, s)=="true")
			bus_on_wire="true";
		# second bus point on segment
		if (point_on_seg(bus_x1[n], bus_y1[n], w, s)=="true")
			bus_on_wire="true";

		tap_on_wire="false";
		# first bus point on segment
		if (point_on_seg(tap_x0[n], tap_y0[n], w, s)=="true")
			tap_on_wire="true";
		# second bus point on segment
		if (point_on_seg(tap_x1[n], tap_y1[n], w, s)=="true")
			tap_on_wire="true";

		if (bus_on_wire=="true")
		     {
			bus_id[n]=id[w];
	 	    }

		if (tap_on_wire=="true")
		     {
			tap_id[n]=id[w];
	 	     }

	  }
      }
   split(wire_type[bus_id[n]],type_base,"(");
   for (v=0; v<wirenum; v++)  
      if (id[v]==tap_id[n]) 
	{
	 #probably unnecessary, either driven: type already known,
	 #or driving: declared via bus declaration
	 #bidir:can be treated as driving
	 #wire_type[v]=type_base[1] tap_range[n];
	 if (tap_dir[n]=="bus_out") 
	    { 
		wire_name[v]=wire_name[bus_id[n]] tap_range[n];
		wire_is_subnet[v]="true";
	    }
	}
  }
}
# identifying generic pins which have init values assigned to them

END \
{
for (c in component)
  {
   for (n=0; n < gpinnum[c]; n++)
     { 
      new_gpin_init[c, n] = gpin_init[c, n];
      # check whether generic pin is connected to a init tag
      for (i=0; i < i_valnum; i++)
         {
          connected = point_in_pin(i_val_x[i],i_val_y[i], \
			gpx[c,n,0],gpy[c,n,0], gpx[c,n,1],gpy[c,n,1], \
			gpx[c,n,2],gpy[c,n,2], gpx[c,n,3],gpy[c,n,3] );
          if (connected == "true") new_gpin_init[c, n] = i_val[i];
         }
     }
  }
}


# processing FSM information
#----------------------------



# determine the net to which an object pin is connected

function find_net_id(ox1,oy1, ox2,oy2, \
		     w1, connected_net)
  {
   # object inputs/outputs are stored as wires as well
   # => there must be a wire which has an input/output point
   #    as a starting point
   connected_net="unconnected";
   for (w1=0; w1<wirenum-1; w1++)
     {
      if( (ox1==wx[w1,0] && oy1==wy[w1,0]) || (ox2==wx[w1,0] && oy2==wy[w1,0]) )
        {
         connected_net=id[w1];
         break;
        }
     }
   return connected_net;
  }

function find_lbl_net_id(lx,ly, \
		         w1, connected_net)
  {
   # check for wire endpoints which are connected to the label
   connected_net="unconnected";
   for (w1=0; w1<wirenum-1; w1++)
     {
      if( (lx==wx[w1,0] && ly==wy[w1,0]) || \
          (lx==wx[w1,points[w1]-1] && ly==wy[w1,points[w1]-1]) )
        {
         connected_net=id[w1];
         break;
        }
     }
   return connected_net;
  }

END {

# find nets which are connected to state symbols

for(s=0;s<statenum;s++)
  {
   state_in_net[s]= \
         find_net_id(state_in_x1[s],state_in_y1[s], \
	             state_in_x2[s],state_in_y2[s]);
   state_out_net[s]= \
         find_net_id(state_out_x1[s],state_out_y1[s], \
	             state_out_x2[s],state_out_y2[s]);
  }


# find nets which are connected to action symbols

for(a=0;a<actnum;a++)
  {
   act_in_net[a]= \
         find_net_id(act_in_x1[a],act_in_y1[a], \
	             act_in_x2[a],act_in_y2[a]);
   act_out_net[a]= \
         find_net_id(act_out_x1[a],act_out_y1[a], \
	             act_out_x2[a],act_out_y2[a]);
  }


# find nets which are connected to decision symbols

for(d=0;d<decnum;d++)
  {
   dec_in_net[d]= \
         find_net_id(dec_in_x1[d],dec_in_y1[d], \
	             dec_in_x2[d],dec_in_y2[d]);
   dec_out_net[d]= \
         find_net_id(dec_out_x1[d],dec_out_y1[d], \
	             dec_out_x2[d],dec_out_y2[d]);
  }


# determine the net id for each label

for(l=0;l<fsmlabelnum;l++)
  {
   # look for nets which are connected to one endpoint
   lbl_net1[l]= find_lbl_net_id(lbl_x1[l],lbl_y1[l]);
   # look for nets which are connected to the other endpoint
   lbl_net2[l]= find_lbl_net_id(lbl_x2[l],lbl_y2[l]);
  }


# find the reset state
for(s=0;s<statenum;s++)
  {
  #print "state " s;
  if(point_on_line(start_x[0], start_y[0], \
		    state_rline_x0[s], state_rline_y0[s], \
                    state_rline_x1[s], state_rline_y1[s])=="true")
       reset_state=state_name[s];
  }

}



# generate the FSM transition process 
#
function gen_trans_proc(net_id, indent, \
		     a,d,s,l,m,first_label)
 {

   for(a=0;a<actnum;a++)
     {
      if(act_in_net[a]==net_id) # connected to an action block
        {
	 # build action
         #*for(l=0;l<act_vhdl_lines[a];l++)
	 #*  print indent act_vhdl[a, l];
	 gsub("\n", "\n" indent, act_vhdl[a]);
	 print indent act_vhdl[a];
	 # continue looking from action block output
	 #*print "connected to action " a;          
	 gen_trans_proc(act_out_net[a], indent);
         return; # no need to continue looking for the old net
        } 
     }
    
   for(d=0;d<decnum;d++)
     {
      if(dec_in_net[d]==net_id) # connected to a decision block
        {
	 for(l=0;l<fsmlabelnum;l++)
	   {
	    #*print "connected to decision " d;          
	    if((lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d]) \
	       && match(tolower(lbl_expr[l]),"else") )
	      {
		first_label="true";
		# this is a decision node with an else clause
	 	for(m=0;m<fsmlabelnum;m++)
		   {
		    if((lbl_net1[m]==dec_out_net[d] || \
		        lbl_net2[m]==dec_out_net[d]) && m != l)
		      {
		       # build the condition which follows this branch
		       printf indent;
		       if(first_label=="false") printf "els";
		       first_label="false";
		       print "if " dec_expr[d] " " lbl_expr[m] " then";

		       # find the actions which occur in here
		       # (nested IFs + state assignments)
	
		       if(lbl_net1[m]==dec_out_net[d])
			  gen_trans_proc(lbl_net2[m], indent "   ");
		       if(lbl_net2[m]==dec_out_net[d]) 
			  gen_trans_proc(lbl_net1[m], indent "   ");

		      }
		   }
	        # else clause
	        print indent "else";
	        # find the actions which occur in here
	        # (nested IFs + state assignments)
	        if(lbl_net1[l]==dec_out_net[d]) 
		   gen_trans_proc(lbl_net2[l], indent "   ");
	        if(lbl_net2[l]==dec_out_net[d]) 
                   gen_trans_proc(lbl_net1[l], indent "   ");

	        # end of the IF statement
	        print indent "end if;";
	     
	        return; # finished this level
	      }
	   }
	 for(l=0;l<fsmlabelnum;l++)
	   {
	    if(lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d])
	      {
	       #*print "connected to label " l;          
	       # build the condition which follows this branch
	       printf indent "if " dec_expr[d] " " lbl_expr[l];
               print " then";

	       # find the actions which occur in here
	       # (nested IFs + state assignments)

	       if(lbl_net1[l]==dec_out_net[d]) 
		  gen_trans_proc(lbl_net2[l], indent "   ");
	       if(lbl_net2[l]==dec_out_net[d]) 
		  gen_trans_proc(lbl_net1[l], indent "   ");

	       # end of the IF statement
               print indent "end if;";
	      }
	   }
         return; # no need to continue looking for the old net
        } 
     }

   for(s=0;s<statenum;s++)
     {
      if(state_in_net[s]==net_id) # found end of branch
        {
         print indent "next_" fsm_name[0] "_state <= " state_name[s] ";";
         return;
        }
     }

   return;
}






# generate the Flow diagram function 
#
function gen_flowdiag(net_id, indent, \
		     a,d,s,l,m,first_label)
 {

   for(a=0;a<actnum;a++)
     {
      if(act_in_net[a]==net_id) # connected to an action block
        {
	 # build action
	 gsub("\n", "\n" indent, act_vhdl[a]);
	 print indent act_vhdl[a];
	 # continue looking from action block output
	 gen_flowdiag(act_out_net[a], indent);
         return; # no need to continue looking for the old net
        } 
     }
    
   for(d=0;d<decnum;d++)
     {
      if(dec_in_net[d]==net_id) # connected to a decision block
        {
	 for(l=0;l<fsmlabelnum;l++)
	   {
	    if((lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d]) \
	       && match(tolower(lbl_expr[l]),"else") )
	      {
		first_label="true";
		# this is a decision node with an else clause
	 	for(m=0;m<fsmlabelnum;m++)
		   {
		    if((lbl_net1[m]==dec_out_net[d] || \
		        lbl_net2[m]==dec_out_net[d]) && m != l)
		      {
		       # build the condition which follows this branch
		       printf indent;
		       if(first_label=="false") printf "els";
		       first_label="false";
		       print "if " dec_expr[d] " " lbl_expr[m] " then";

		       # find the actions which occur in here
		       # (nested IFs + state assignments)
	
		       if(lbl_net1[m]==dec_out_net[d])
			  gen_flowdiag(lbl_net2[m], indent "   ");
		       if(lbl_net2[m]==dec_out_net[d]) 
			  gen_flowdiag(lbl_net1[m], indent "   ");

		      }
		   }
	        # else clause
	        print indent "else";
	        # find the actions which occur in here
	        # (nested IFs + state assignments)
	        if(lbl_net1[l]==dec_out_net[d]) 
		   gen_flowdiag(lbl_net2[l], indent "   ");
	        if(lbl_net2[l]==dec_out_net[d]) 
                   gen_flowdiag(lbl_net1[l], indent "   ");
	     
	        return; # finished this level
	      }
	   }
	 for(l=0;l<fsmlabelnum;l++)
	   {
	    if(lbl_net1[l]==dec_out_net[d] || lbl_net2[l]==dec_out_net[d])
	      {
	       # build the condition which follows this branch
	       printf indent "if " dec_expr[d] " " lbl_expr[l];
               print " then";

	       # find the actions which occur in here
	       # (nested IFs + state assignments)

	       if(lbl_net1[l]==dec_out_net[d]) 
		  gen_flowdiag(lbl_net2[l], indent "   ");
	       if(lbl_net2[l]==dec_out_net[d]) 
		  gen_flowdiag(lbl_net1[l], indent "   ");

	      }
	   }
         return; # no need to continue looking for the old net
        } 
     }

   for(s=0;s<statenum;s++)
     {
      if(state_in_net[s]==net_id) # found end of branch
        {
         print indent "next_" fsm_name[0] "_state <= " state_name[s] ";";
         return;
        }
     }

   for(j=0;j<joinnum;j++)
     {
      if(join_in1_net[j]==net_id) # this is the end of a main branch
        {
         print indent "end if;";
         return;
        }
      if(join_in2_net[j]==net_id) return; # end of side branch
     }

   return;
}




# generating VHDL code
#----------------------


END \
{
	print "--";
	print "-- file generated by VGI";
	#system("echo -n \"-- at\"; date");
	system("echo \"-- at\" `date` ");
	print "--";
}

END		{print ""}
END		{print ""}
END		{system("touch .vgi_use_clauses; cat .vgi_use_clauses")}
END		{print ""}
END		{print ""}


END		{print "entity " entity " is"}
END		{if (gportnum > 0)
		  {
		   print "\tgeneric";
		   print "\t    (";
		   for (n=0; n < gportnum; n++)
		     { 
		       line="\t\t" gport_name[n] ":\t" gport_type[n];
		       if (gport_init[n]!=" ") line=line "\t:= " gport_init[n];
		       if (n < gportnum-1) line=line ";";
		       print line;
		     }
		   print "\t    );";
		  }
		}
END		{if (portnum > 0)
		  {
		   print "\tport( ";
		   for (n=0; n < portnum; n++)
		     { if (n < portnum-1)
			print "\t\t" port_name[n] ":\t" port_dir[n] "\t" \
			port_type[n] ";";
		       else
			print "\t\t" port_name[n] ":\t" port_dir[n] "\t" \
			port_type[n]
		     }
		   print "\t    ); ";
		  }
		}
END		{print "end " entity ";"}
END		{print ""}
END		{print ""}


END		{print "architecture " arch " of " entity " is"}
END		{print ""}
END		{print ""}

#
# signal declarations
#

END \
{
 for (w=0; w < wirenum; w++)
    {
	# find out whether signal has been declared already
	already_declared="false";
	for (pw=w-1; pw >= 0; pw--)
	   { if (wire_name[w]==wire_name[pw]) already_declared="true";}
	for (p=0; p < portnum; p++)
	   { if (wire_name[w]==port_name[p] && (port_dir[p]=="in" || \
                 port_dir[p]=="inout"))
		already_declared="true";}
	if (already_declared=="false" && wire_is_subnet[w]=="false" && \
	    wire_type[w]!="#none#")
	   { print "signal " wire_name[w] ":\t" wire_type[w] ";"; }
    }
 print "";
 print "";
}
END		{print "-- vgi_off (User defined signals go here)"}
END		{system("touch .vgi.tmp.user_signals; cat .vgi.tmp.user_signals")}
END		{print "-- vgi_on"}
END		{print ""}


#
# component declarations
# 
END		{for (c in component)
		   {
		    if (new_component[c]=="true")
		     {
			print "component " component[c];
			{if (gpinnum[c] > 0)
			  {
			   print "\tgeneric";
			   print "\t    (";
			   for (n=0; n < gpinnum[c]; n++)
			     { 
			      line="\t\t" gpin_name[c, n] ":\t" gpin_type[c, n];
			      if (gpin_init[c, n]!=" ") line=line "\t:= " \
						        gpin_init[c, n];
			      if (n < gpinnum[c]-1) line=line ";";
			      print line;
			     }
			   print "\t    );";
			  }
			}
			if(pinnum[c]>0)
			 {
			  print "\tport( ";
			  for (n=0; n < pinnum[c]; n++)
			     {
			      if (n < pinnum[c]-1)
				print "\t\t" pin_name[c, n] ":\t" \
				pin_dir[c, n] "\t" \
				pin_type[c, n] ";";
			       else
				print "\t\t" pin_name[c, n] ":\t" \
				pin_dir[c, n] "\t" \
				pin_type[c, n] \
			     };
			  print "\t    );";
			 };
			print "end component;";
			print "";
			print "";
		     } \
		   } \
		}
END		{print ""}
END		{print ""}



#
# FSM declarations
#

END {
if(statenum>0)
 {
  # extract state type

  printf "  type " fsm_name[0] "_type is (";
  for(s=0;s<statenum;s++)
    {
     printf state_name[s];
     if(s!=statenum-1) printf ", ";
    }
  print ");"
  print "";
  print "  signal " fsm_name[0] "_state: " fsm_name[0] "_type;";
  print "  signal next_" fsm_name[0] "_state: " fsm_name[0] "_type;";
  print "";
  print "";
  print "";
}}


END		{print "begin"}
END		{print ""}
END		{print ""}

#
# component instantiations
#


END		{comp_id=0}
#*END		#*{for (c in component)
END		{for (c=0; c < comnum; c++)
		   {
			#*print "U" comp_id ": "component[c];
			#*comp_id++;
			print "U" c ": "component[c];
			{if (gpinnum[c] > 0)
			  {
			   print "\tgeneric map";
			   print "\t    (";
			   for (n=0; n < gpinnum[c]; n++)
			     { 
			      line="\t\t" gpin_name[c, n];
			      if (gpin_init[c, n]!=" ") line=line "\t=> " \
						        new_gpin_init[c, n];
			      if (n < gpinnum[c]-1) line=line ",";
			      print line;
			     }
			   print "\t    )";
			  }
			}
			print "\tport map( ";
			for (n=0; n < pinnum[c]; n++)
			     {
			      #print pin_id[c, n];
			      #print wire_name[pin_id[c, n]];
			      if (pin_id[c, n] < 0)
				{actual="open"}
			       else
				{actual=wire_name[pin_id[c, n]]}
			      if (n < pinnum[c]-1)
				print "\t\t" pin_name[c, n] "\t" \
				"=> " actual ",";
			       else \
				print "\t\t" pin_name[c, n] "\t" \
				"=> " actual;
			     };
			print "\t\t);";
			print "";
			print "";
		   }
		}

# connecting the symbol outputs/bidirs to their internal signals
END		{
		#print "###==================";
		#print "### portnum=" portnum;
		 if (portnum>0)
		  { 
		    print "-- connecting internal signals to their entity ports\n";
		  }; 
		 for (n=0; n < portnum; n++)
		  if (port_dir[n]=="out")
		   {
		     if (port_id[n]>0)
		   	#print "### port=" n;
			#print port_name[n] "\t<= i_" port_name[n] ";";
		   	print port_name[n] "\t<= " wire_name[port_id[n]] ";";
		   }
		}
END		{print ""}
END		{print ""}


# connecting busses
END		{
		 if (tapnum>0) print "-- connecting busses\n";
		 for (n=0; n < tapnum; n++)
		  if (tap_dir[n]=="bus_in" || tap_dir[n]=="bus_inout")
		   {
		     print wire_name[bus_id[n]] tap_range[n] \
			   "\t<= " wire_name[tap_id[n]] ";";
		   }
		  if (tap_dir[n]=="bus_inout")
		   {
		     print wire_name[tap_id[n]] \
			   "\t<= " wire_name[bus_id[n]] tap_range[n] ";";
		   }
		}
END		{print ""}
END		{print ""}


# insert truth tables
#---------------------

function xp_iv(in_vars, xpos, out_vars,   v, v0 ) {

   for(v=xpos;v<=length(in_vars);v++) {
	if(substr(in_vars,v,1)=="-") {
	   v0=in_vars;
	   sub("-","0",v0);
	   sub("-","1",in_vars);
	   xp_iv(v0,v+1,out_vars);
          }
     }
    case_line[in_vars] = "      when " "\"" in_vars "\" => ";
    for(v=1;v<=length(out_vars);v++) {
        case_line[in_vars] = case_line[in_vars] ttab_res[v] \
                          " <= '" substr(out_vars,v,1) "'; ";
      }
    return 0;
}


END {
for(ttab=0;ttab<ttabnum;ttab++) {
   #print "####" ttablines[ttab];
   #for(tl=0;tl<ttablines[ttab];tl++) print truthline[ttab,tl];

   for(tl=0;tl<ttablines[ttab];tl++) 
   if(match(truthline[ttab,tl],"[[]")) # found header
     {
      gsub("[[]"," ",truthline[ttab,tl]);
      gsub("[]]"," ",truthline[ttab,tl]);
      split(truthline[ttab,tl],sigs,"[|]");
      ttsig_num = split(sigs[1],ttab_sigs," ");
      ttres_num = split(sigs[2],ttab_res," ");
      print "";
      printf "-- ";
      for(i=1;i<=ttres_num;i++) printf ttab_res[i] " ";
      print "- Truth table";
      print "";
      printf "ttab" ttab ": process(";
      for(s=1; s<ttsig_num; s++)
         printf ttab_sigs[s] ", ";
      print ttab_sigs[ttsig_num] ")";
      print "";
      printf "   variable ttab_val: std_logic_vector(";
      printf ttsig_num-1;
      print " downto 0);";
      print "";
      print "begin";
      print "";
      printf "   ttab_val := ";
      for(s=1; s<ttsig_num; s++)
         printf ttab_sigs[s] "& ";
      print ttab_sigs[ttsig_num] ";";
      print "";
      print "   case ttab_val is";
      delete case_line;
      for(l=tl+1;l<ttablines[ttab];l++) 
        if(!match(truthline[ttab,l],"--")  && \
	   (match(truthline[ttab,l],"-") ||  \
	    match(truthline[ttab,l],"1") ||  \
	    match(truthline[ttab,l],"0")  )  ){
   	   gsub("[\*]","-",truthline[ttab,l]);
           split(truthline[ttab,l],vals,"[|]");
           gsub(" ","",vals[1]);
           gsub(" ","",vals[2]);
           # expand irrelevant variables
	   xp_iv(vals[1], 1, vals[2]);
        }
      for(c in case_line) print case_line[c];
      print "      when others =>";
      print "   end case;";
      print "";
      print "end process ttab" ttab ";";
      print "";
      break;
     }
   }
}



# insert state machines
#-----------------------

END \
{
if(statenum>0)
 {
  print "";
  print "--";
  print "-- FSM: ";
  print "-- " fsm_name[0];
  print "--";
  print "";

  print "";
  print "-- state register";
  print "-----------------";
  print "";
  print fsm_name[0] "_reg: process(" fsm_clk_name "," fsm_rst_name ")";
  print "";
  print "";
  print "begin";
  print "";

  # use a synchronous reset
  if(fsm_sync_reset=="true") {
      print "   if " fsm_clk " then";
      print "";
      print "      if " fsm_rst " then";
      print "";
      print "         " fsm_name[0] "_state <= " reset_state ";";
      print "";
      print "      end if;";
      print "";


     }
   # use an asynchronous reset
   else { 
      print "   if " fsm_rst " then";
      print "";
      print "      " fsm_name[0] "_state <= " reset_state ";";
      print "";
      print "   elsif " fsm_clk " then";
    }

  print "";
  print "      " fsm_name[0] "_state <= next_" fsm_name[0] "_state;";
  print "";
  print "   end if;";
  print "";
  print "end process;";
  print "";
  print "";



  # generate the transition/output process

  print "";
  print "-- transition/output process";
  print "----------------------------";
  print "";
  printf fsm_name[0] "_trans: process("
  for (p=0; p < portnum; p++)
    {if(port_dir[p]=="in" || port_dir[p]=="inout")
	printf port_name[p] ", ";
     }
  printf fsm_name[0] "_state)";
  print "";
  print "";
  print "";
  print "begin";
  print "";
  print fsm_y0;
  print "";
  print "      case " fsm_name[0] "_state is";

  for(s=0;s<statenum;s++)
    {
     print "";
     printf "         when ";
     printf state_name[s];
     print " =>";
     print "";
     gen_trans_proc(state_out_net[s], "\t\t");
    }
  print "";

  print "      end case;";

  print "";
  print "end process;";
  print "";
  print "";
 }
}





END		{print "-- vgi_off (User defined code goes here)"}
#END		{system("awk -f .vgi.tmp.user_code x")}
END		{system("touch .vgi.tmp.user_code; cat .vgi.tmp.user_code")}
END		{print "-- vgi_on"}
END		{print ""}

END		{print "end " arch ";"}


############ debugging stuff


END		{
	debug_mode="off";
	if (debug_mode=="on")
	    {
		 for (n=0; n < wirenum; n++)
		 { 
			print "wirenr: " n " name: " wire_name[n];
			print "id: " id[n] " subnet=" wire_is_subnet[n];
		   print "points: " points[n];
		   print "id: " id[n];
		   for (m=0; m < points[n]; m++)
		     {
			print wx[n, m] "," wy[n, m];
		     }
		 }
		
		print "labelnum=" labelnum;
		 for (n=0; n < labelnum; n++)
		   {
			print  "labelnr: " n " labelname " labelname[n] " \
				labeltype " labeltype[n];
			print "label_id: " label_id[n];
			print "lx0: " lx0[n] " ly0: " ly0[n] " lx1: " \
				lx1[n] " ly1: " ly1[n];
		   }

		print tapnum;
		 for (n=0; n < tapnum; n++)
		   {
			print "tapnr: " n " tap_range=" tap_range[n];
			print "bus_id=" bus_id[n] " tap_id=" tap_id[n];
		   }
		print "comnum=" comnum; 
		for (c=0; c < comnum; c++)
		  {
		    print "com=" c; 
		    for (p=0; p < pinnum[c]; p++)
		      {
		      print "pin=" p " " pin_dir[c,p]; 
		      for (n=0; n < 6; n++)
			{
			 print "x,y=" ipx[c,p,n] "," ipy[c,p,n]; 
			}
	   	      }
		  }


		print "valnum=" i_valnum; 
		for (v=0; v < i_valnum; v++)
		  {
		    print "val=" i_val[v]; 
		    print "x=" i_val_x[v];
		    print "y=" i_val_y[v];
		  }



	    }
}


# FSM debugging stuff
#####################

END {
debug="false";
if(debug=="true")
 {
     # print Moore states

     # print the action which is connected with the current state
     for(a=0;a<actnum;a++)
       if(state_out_net[s]==act_in_net[a])
         {
          for(l=0;l<act_vhdl_lines[a];l++) 
              print "\t    " act_vhdl[a,l];
         }
     print "";

     for(d=0;d<decnum;d++) 
       {
        printf "\t    if " dec_expr[d];
        print " then";
        print "\t    ";
        print "\t    end if";
       }
  for(s=0;s<statenum;s++) 
   {
    print state_name[s] ":";
    print "in=" state_in_net[s];
    print "out=" state_out_net[s];
    print state_in_x1[s] ", " state_in_y1[s] " / " \
          state_in_x2[s] ", " state_in_y2[s];
    print state_out_x1[s] ", " state_out_y1[s] " / " \
          state_out_x2[s] ", " state_out_y2[s];

   }
  for(l=0;l<fsmlabelnum;l++) 
   {
    print lbl_expr[l] ":";
    print "net1=" lbl_net1[l];
    print "net2=" lbl_net2[l];
    print lbl_x1[l] ", " lbl_y1[l] " / " \
          lbl_x2[l] ", " lbl_y2[l];
   }
  for(a=0;a<actnum;a++) 
   {
    print "action " a ":";
    print "in=" act_in_net[a];
    print "out=" act_out_net[a];
   }
  for(d=0;d<decnum;d++) 
   {
    print "decision " d ":";
    print "in=" dec_in_net[d];
    print "out=" dec_out_net[d];
   }
		 for (n=0; n < wirenum; n++)
		 { 
			print "wirenr: " n " name: " wire_name[n];
			print "id: " id[n] " subnet=" wire_is_subnet[n];
		   print "points: " points[n];
		   print "id: " id[n];
		   for (m=0; m < points[n]; m++)
		     {
			print wx[n, m] "," wy[n, m];
		     }
		 }

 }

}
